/* PK.C
	Point kinetics analysis for the MNR power excursion incident
  version: TEST.C
			  - changed printout and added an input routine
			  - changed the initial n calculation
  version PKSIN.C
			 - added flux weighted control rod worth
  version PKSIN1.C
			 - added flux weighted fuel assembly worth
  version PKSIN2.C
			 - added flux**2 weighted worths - full 1/2 cosine
  version PK3.C
			 - flux**2 weighted worths - partial 1/2 cosine
  version PK4.C
			 - addition printout
			 - adjusted time step control
  version PK5.C
			 - parameters read in by input rather than hard coded:
							  beta correction factor
							  rod drop delay
							  prompt neutron lifetime
  version PKSIM.C
			 - General version for LOR simulations

*/
#define TRUE 1
#define FALSE 0
#define PI 3.14159

#include <stdio.h>
#include <math.h>
#include <process.h>
#include <stdlib.h>
#include <string.h>

void calc_dk_add(void);
void calc_dk_trip(void);
void calc_dk_regrod(void);
void input(void);
void output_head(void);
void output_result(void);

  FILE *fp, *fp1;
  char buf[80];
  float lambda[6], betasum, beta[6], betafactor, life;
  float n, nold, dndt, c[6], dc[6], sum, Po, P, Pscale;
  float k, k_init, t_init, k_fin, t_fin,dk_add,dk_reg; //reactivity insertion
  float t, dtmin, dtnominal, dtmax, dt, tstart, tfin, tolerance, tprint, dtprint;
  float Pmax=0.0, tmaxpower, tau, integral = 0.0;
  float height, y, x, rod_worth, dk_trip, t_trip, t_delay; //SDS parameters
  float P_trip, k_trip, P_tripsignal, t_tripsignal, k_tripsignal;
  float distance, print_trip, print_drop, highpower_trip;
  int i, flx_wt_rod = 0, TRIP = FALSE;


int main( int argc, char *argv[] )
{
  input(); //Read input data

  dt = dtnominal; //initial time step (seconds)
  t = 0.0;   //start time (seconds)
  tprint = 0.0;
  dk_trip = 0.0; //initially not tripped
  t_trip = tfin +1.;
  print_trip = TRUE;
  print_drop = TRUE;
  distance = 0.0;
  dk_trip = 0.0;
  tau = 9.9e+9;
  k = k_init;
  P = n * Pscale;

  betasum = .00695*betafactor;
  lambda[0] = .0127;
  lambda[1] = .0317;
  lambda[2] = .1155;
  lambda[3] = .3108;
  lambda[4] = 1.3974;
  lambda[5] = 3.8723;
  beta[0] = 0.038 * betasum;
  beta[1] = 0.213 * betasum;
  beta[2] = 0.188 * betasum;
  beta[3] = 0.407 * betasum;
  beta[4] = 0.128 * betasum;
  beta[5] = 0.026 * betasum;
  //assume precursors at equilibrium at start
  for (i=0; i<6; i++) c[i] = beta[i]*k*n/life/lambda[i];

  output_head();  //print headers in output files
  output_result(); //print output data at initial state

  while (t <= tfin)
  {
	 t = t + dt;
	 //reactivity change
	 calc_dk_add();  //update manual delta k insertion

	 calc_dk_trip(); // update shutdown system induced delta k

	 calc_dk_regrod(); // update control system induced delta k

	 k = k_init + dk_add + dk_trip + dk_reg;

	 //update neutron and precursor levels
	 sum = 0.0;
	 for (i=0; i<6; i++) sum = sum + lambda[i] * c[i];
	 //Explicit method
	 //dndt = (k*(1-betasum)-1)/life * n  + sum;
	 //dn = dndt*dt;
	 //n = n + dn;
	 //Semi-implicit method
	 nold = n;
	 n = (n + dt*sum) / (1-dt*(k*(1-betasum)-1)/life);
	 dndt = (n-nold)/dt;
	 tau = 9.9e+9;
	 if (n > nold) tau = dt/log(n/nold);
	 if (n < nold) tau = -dt/log(nold/n);
	 for (i=0; i<6; i++)
	 {
		//Explicit method
		//dc[i] = (beta[i]*k/life * n - lambda[i] * c[i]) * dt;
		//c[i] = c[i] + dc[i];
		//Semi-implicit method
		c[i] = (c[i] +  dt* beta[i]*k/life * n) / (1.0+lambda[i]*dt);
	 }
	 P = n * Pscale;
	 if (P > Pmax)
	 {
		Pmax = P;
		tmaxpower = t;
	 }
	 if (t >= tprint)
	 {
		tprint = t + dtprint;
		output_result();
	 }

	 //adjust dt
	 dt = dtnominal;
	 //   if (k > 1.0 && dndt != 0.0) dt = tolerance*n / fabs(dndt);
	 if (dndt != 0.0) dt = tolerance*n / fabs(dndt);
	 if (dndt == 0.0) dt = dtmax;
	 if (dt < dtmin) dt = dtmin;
	 if (dt > dtmax) dt = dtmax;
  }
  printf("Trip signal: Power = %g at %g seconds, k = %g\n",
			 P_tripsignal, t_tripsignal, k_tripsignal);
  fprintf(fp,"Trip signal: Power = %g at %g seconds, k = %g\n",
				  P_tripsignal, t_tripsignal, k_tripsignal);
  printf("Trip bite: Power = %g at %g seconds, k = %g\n",
			 P_trip, t_trip, k_trip);
  fprintf(fp,"Trip bite: Power = %g at %g seconds, k = %g\n",
				  P_trip, t_trip, k_trip);
  printf("maximum power of %g at %g seconds",Pmax, tmaxpower);
  fprintf(fp,"maximum power of %g at %g seconds",Pmax, tmaxpower);
  fclose (fp);
  fclose (fp1);
  return;
}

//---------------calc_dk()---------------------------------
void calc_dk_add()
{
  float tau_sp, step;
  tau_sp=0.0;

  //pure linear addition
  //dk_add = 0.0;
  //if (t > t_init) dk_add = (k_fin-k_init)/(t_fin-t_init)*(t-t_init);

  //linear addition with hold if period < given setpoint
  if (t > t_init)
  {
	 step  = (k_fin-k_init)/(t_fin-t_init)*dt;
	 if ( abs(tau) < abs (tau_sp)) step = 0.0;
	 dk_add = dk_add+step;
  }
  //clip the addition to the maximum allowed
  if (dk_add > (k_fin-k_init)) dk_add = k_fin-k_init;
}

void calc_dk_trip()
{
  //Test for high power trip condition
  if (P >= highpower_trip*Pscale) TRIP = TRUE;
  if (TRIP == TRUE)
  {
	 if (print_trip)
	 {
		t_trip = t + t_delay;
		t_tripsignal = t;
		P_tripsignal = P;
		k_tripsignal = k;
		printf("trip signal at:\n");
		fprintf(fp,"trip signal at:\n");
		print_trip = FALSE;
		output_result();
	 }
	 if (t >= t_trip)
	 {
		if (print_drop)
		{
		  P_trip = P;
		  k_trip = k;
		  printf("rod dropped and starting to bite at:\n");
		  fprintf(fp,"rod dropped and starting to bite at:\n");
		  output_result();
		}
		print_drop = FALSE;
		distance = 0.5 * 560.0 * (t-t_trip)*(t-t_trip);
		//distance in cm. of rod drop
		height = 60.0;
		rod_worth = -0.0782;
		if (distance > height) distance = height;
		y = (-25.+ distance)/70.0*PI;
		if (flx_wt_rod == 1)
		  dk_trip = rod_worth/3.0837*(y+1.1220+sin(y)*cos(y)+0.3909);
		else dk_trip = rod_worth*distance/height;
		if (dk_trip < rod_worth) dk_trip = rod_worth; //max rod worth
	 }
  }
}

void calc_dk_regrod()
{
float reg_height, regrod_worth, gain, taud, taui, offset, regrod_pos, reg_y;
int AUTOREG = TRUE, flx_wt_reg = 0;
  //Test that reg rod controller is active
//  AUTOREG == FALSE;
  dk_reg = 0.0;
  if (AUTOREG == TRUE)
  {
		reg_height = 60.0;
		regrod_worth = -0.006;
		gain = .2;
		taud = 3.8; // seconds
		taui = 1.0/0.7; //seconds
		offset = 30.0;
		integral = integral + (n-1.0)*dt;
		//demand insertion of regrod (cm.)
		regrod_pos = gain*(n-1.0 + taud*dndt +
						integral/taui)+offset;//PI controller
		if (regrod_pos > reg_height) regrod_pos = reg_height;
		if (regrod_pos < 0.0) regrod_pos = 0.0;
		reg_y = (-25.+ regrod_pos)/70.0*PI;
		if (flx_wt_reg == 1)
		  dk_reg = regrod_worth/3.0837*(reg_y+1.1220+sin(reg_y)*cos(reg_y)
						+0.3909);
		else dk_reg = regrod_worth*(regrod_pos-reg_height/2.0)/reg_height;
  }
}


void input()
{
  /* Open an existing file for reading. Fail if file doesn't exist. */
	 if( (fp = fopen( "input.inp", "r")) != NULL )
	 {
/*Sample Input for point kinetics run
.000001   // dtmin, minimum time step
.5        // dtmax, maximum time step
.0001     // dtnominal, nominal time step
.0001     // tolerance, tolerance
.01       // dtprint, print increment
25.       // tfin, end time
1.002     // k_fin, final k after manual adition (units of k)
0.        // t_init, start time for k addition  (sec)
1.        // t_fin, end time for k addition (sec)
1.0       // initial k at t=0
1.0       // n, initial value at t=0
2000000.0 // Pscale, nominal power for n=1 (watts)
1         // 1 = flux wt the shim rod
1.0       // beta correction factor
1.25      // high power trip setpoint, fraction
0.025     //rod drop delay (sec)
0.000051  //prompt neutron lifetime
*/
  /* Title card */
		 fgets (buf, 80, fp);
	/* read input data  */
		 fscanf (fp, "%f", &dtmin);  // minimum time step
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &dtmax);  // maximum time step
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &dtnominal);  // nominal time step
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &tolerance);  // tolerance
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &dtprint);  // print increment
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &tfin);  // end time
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &k_fin); //manual delta k addition
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &t_init);  //start time for delta k insertion
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &t_fin);  //end time for delta k insertion
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &k_init);  //initial k
		  fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &n);  //initial n
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &Pscale);  //nomimal power (watts) for n = 1
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%d", &flx_wt_rod);  //flux wt flag
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &betafactor);  //beta correction factor
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &highpower_trip);//high power trip setpoint, fraction
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &t_delay);  //rod drop delay
			fgets (buf, 80, fp); /* read comments */
		 fscanf (fp, "%f", &life);  //prompt neutron lifetime
			fgets (buf, 80, fp); /* read comments */
		fclose( fp );
	 }
	 else	/* File did not exist. Print message */
	 {
		 printf("Could not open the input file");
		 return;
	 }

}

void output_head()
{
	if( (fp = fopen( "out.dat", "w")) == NULL ||
		(fp1 = fopen( "out.prn", "w")) == NULL)
	{
	  /* File could not be opened. Print message */
		printf( "File does not exist\n" );
		return;
	}
	printf("%g %g %g %g %g %g \n",
	 dtmin,dtmax,dtnominal,tolerance,dtprint,tfin);

	fprintf(fp,"%g %g %g %g %g %g \n",
	 dtmin,dtmax,dtnominal,tolerance,dtprint,tfin);

	fprintf(fp,
	"time       dt        Power             n    k     dk_trip  distance tau\n");

}

void output_result()
{
		printf("t=%f,dt=%f,n=%f,k=%f \n", t, dt, n, k);
		fprintf(fp,"%8.6f %8.6f %10.6f %8.5g %7.5f %7.5f %5.3f %7.5g\n",
						t, dt, P, n, k, dk_trip, distance, tau);
		fprintf(fp1,"%10.6f %10.6f\n", t, P);
}